/**
 * Copyright (C) Intelligent Media LTD - All Rights Reserved
 *
 * This source code is protected under international copyright law.  All rights
 * reserved and protected by the copyright holders.
 * This file is confidential and only available to authorized individuals with the
 * permission of the copyright holders.  If you encounter this file and do not have
 * permission, please contact the copyright holders and delete this file.
 *
 * Author : Apostolos Katranitsas <ak@intelligentmedia.gr>
 * Contributors :
 *   -
 */
/*
 * Library Manager
 * 
 * V0.0.1
 */
/* global LibraryManager, DeviceManager, CommunicationManager, defined, ERRORS, Logger, utils */

(function () {
  "use strict";

  /**
   * @namespace LibraryManager
   * @classdesc Provide Library handling.
   * @type LibraryManager
   * @version 1.8.1
   * @since 1.8.1
   */
  //jshint -W057
  window.LibraryManager = new function _LibraryManager() {
    /*
     * File should contain
     * + filename
     * + size
     * + type (from v_Downloadable_Types)
     * + isCorrect
     * + lastChecked
     */

    // <editor-fold defaultstate="collapsed" desc="Abstract libFile">

    /**
     * Library file.
     * @constructor
     * @param {String} id The id of the file.
     * @param {String} filename The filename of the file. This is usually the saveAs property from the CMS.
     * @param {String} size The size of the file.
     * @param {String} type One of [media, layout, resource] as type of the file.
     * @returns {LibraryManager.libFile} A new library file.
     */
    this.libFile = function (id, filename, size, type) {
      var _self = this;
      this.id = id;
      this.filename = filename;
      this.size = size;
      this.type = type;
      this.md5 = null;
      this.isCorrect = false;
      /**
       * Last time file checked. This is updated when add a new file in library and when download it.
       */
      this.lastChecked = Date.now() / 1000 | 0;
      /**
       * Manually invalidate a file.
       * This is usually used for resources that has not updated since lastChecked + (updateInterval | 5 mins).
       * @property {Boolean} LibraryManager.manuallyInvalidated Manually invalidate file.
       * @version 1.8.1
       * @since 1.8.1
       */
      this.manuallyInvalidated = false;

      /**
       * Check if file is valid.
       * @param {Function} cbSuccess Function to call if check is success. Doesn't mean it is valid.
       * @param {Function} cbFailure Function to call if check fails.
       * @param {Object} options Object with md5, size and updated fields to check against.
       * @version 1.8.1
       * @since 1.8.1
       */
      this.validateFile = function (cbSuccess, cbFailure, options) {
        //var options.isCorrect = false;
        var fileOptions = {
          path: this.filename
        };

        var checkMD5Cb = function (cbObject) {
          if (cbObject === options.md5) {
            _self.isCorrect = true;
            cbSuccess({isValid: true});
          } else {
            _self.isCorrect = false;
            cbSuccess({isValid: false, reason: "Wrong MD5!"});
          }
        };

        var checkSizeCb = function (cbObject) {
          if (type === "resource") {
            if (_self.updated === options.updated) {
              _self.isCorrect = true;
              _self.times = {
                atime: new Date(cbObject.atime),
                mtime: new Date(cbObject.mtime),
                ctime: new Date(cbObject.ctime)
              };
              cbSuccess({isValid: true});
            } else {
              _self.isCorrect = false;
              cbSuccess({isValid: false, reason: "Wrong update time!"});
            }
          } else {
            if (cbObject.size === parseInt(options.size)) {
              checkMD5Cb(_self.md5);
            } else {
              /* File exists but has wrong size. Delete it here so it can be downloaded again. */
              var fileOptions = {
                path: filename
              };
              DeviceManager.FileSystem.deleteFile(function () {
                _self.isCorrect = false;
                cbSuccess({
                  isValid: false,
                  reason: "Wrong file size (expected: " + options.size + ", got: " + cbObject.size + ")!"
                });
              }, cbFailure, fileOptions.path);
            }
          }
        };

        var checkExistCb = function (cbObject) {
          if (cbObject.exists === true) {
            DeviceManager.FileSystem.fileSize.call(_self, checkSizeCb, cbFailure, fileOptions);
          } else {
            _self.isCorrect = false;
            cbSuccess({isValid: false, reason: "File does not exist!"});
          }
        };

        if (_self.manuallyInvalidated) {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "Manually invalidated!"});
        } else {
          DeviceManager.FileSystem.fileExists.call(_self, checkExistCb, cbFailure, fileOptions);
        }
      };

      //this.downloader = null;


      try {
        if (defined(window.LibraryManager)) {
          if (defined(window.LibraryManager.blacklisted_widgets)) {
            if (window.LibraryManager.blacklisted_widgets.indexOf(id) > -1) {
              this.isBlacklisted = true;
            }
          }
        }
      } catch (e) {
        console.error('LibraryManager.libFile: New file error', e.message, e.stack);
      }

    };

    this.libFile.prototype.download = function (cbSuccess, cbFailure) {
      //console.debug("Start downloading file! ", _self);
      if (this.downloader) {
        this.downloader(cbSuccess, cbFailure);
      } else if (this._downloader) { //The new schema will use _ as internal functions
        this._downloader(cbSuccess, cbFailure);
      } else {
        console.warn("File has no downloader! ", this);
      }
    };

    this.libFile.prototype.validate = function (cbSuccess, cbFailure, options) {
      //console.debug("Start downloading file! ", _self);
      if (this.validator) {
        this.validator(cbSuccess, cbFailure, options);
      } else if (this._validator) { //The new schema will use _ as internal functions
        this._validator(cbSuccess, cbFailure, options);
      } else {
        console.warn("File has no validator! ", this);
      }
    };

    this.libFile.prototype.httpDownloader = function(cbSuccess, cbFailure){

      var dlInfo = {
        source: this.path,
        destination: this.filename,
        downloadStatus: function (file) {
          var percentage = Math.round((((file.amountReceived / file.amountTotal) * 100 + Number.EPSILON) * 10)) / 10;
          percentage = file.status === 'completed' ? '' : percentage;
          application.statusPage.downloadStatus(percentage);
          application.statusPage.downloadStatus.notifySubscribers();
        }
      };

      var _self = this;
      DeviceManager.FileSystem.downloadFile(function (cbobject) {
        _self.isCorrect = true;
        var options = {};
        _self.validate(function (cbobject) {
          _self.lastChecked = utils.getTimeStamp();
          if (cbobject.hasOwnProperty('isValid')) {
            if (!cbobject.isValid) {
              Logger.cmsLog({
                category: "audit",
                message: "Validating media " + _self.filename + " failed. Reason: " + cbobject.reason,
                mediaID: _self.id,
                method: "validateFile",
                thread: "LibraryManager.addFile.downloader"
              });
            }
          }
          try {
            if (_self.filename.endsWith(".htz")) {
              var from = _self.filename;
              var to = _self.filename.substr(0, _self.filename.length - 4);
              DeviceManager.FileSystem.unzipFile(
                function () {
                  cbSuccess(cbobject);
                },
                function () {
                },
                {from: from, to: to});
            } else {
              cbSuccess(cbobject);
            }
          } catch (e) {
            //Logger.cmsLog({category:"error", message:"Could not remove file/dir with name: "+fe.filename+" with error:"+JSON.stringify(e), method:"downloader", thread:"LibraryManager.addFile"});
            // XXX: make a suggestion for a proper error code for this case
            Logger.cmsFault(undefined, undefined,
              {
                code: ERRORS.FaultCodes.RemoteResourceFailed,
                reason: "Could not remove file/dir with name: " + fe.filename + " with error:" + JSON.stringify(e),
                method: "downloader",
                thread: "LibraryManager.addFile",
                date: new Date(),
                expires: new Date((new Date()).valueOf() + DeviceManager.Config.getOptions().collectInterval * 1000)
              }
            );
            if (cbFailure) {
              cbFailure();
            }
          }
        }, function (cbObj) {
          Logger.cmsLog({
            category: "audit",
            message: "Download file failed (not valid). Reason: " + JSON.stringify(cbObj),
            method: "validateFile",
            thread: "LibraryManager.addFile.downloader"
          });
          cbFailure(cbObj);
        }, options);
      }, function (cbObject) {
        var message = JSON.stringify(cbObject);
        if (cbObject.httpStatus && cbObject.httpStatus === 404) {
          message = "File not found. Widget id: " + fe.id;
        }
        ;
        Logger.cmsLog({
          category: "audit",
          message: "Monitor download file failed. Reason: " + message,
          method: "validateFile",
          thread: "LibraryManager.addFile.downloader"
        });
        cbFailure(cbObject);
      }, dlInfo);
    };

    /**
     * Return a file from localStorage entries
     * @static
     * @param file A file from localStorage
     * @return {Layout | Media | Dependency | Widget | Resource} An appropriate file bases on its type.
     */
    this.libFile.factoryFromLocalStorage = function(file){
      file.download = file.downloadFrom;
      var fe;
      if (file.type === Downloadable_Type_Dependency) {
        fe = Dependency.parse(file);
      } else if (file.type === Downloadable_Type_Resource ){
        fe = Resource.parse(file);
      } else if (file.type === Downloadable_Type_Widget ){
        var fileName = file.id+".json";
        var fe = window.LibraryManager.getFileByName(fileName); //XXX: Check that this is valid!
        fe = Widget.parse(file);
      } else if (file.type === Downloadable_Type_Layout ){
        fe = Layout.parse(file);
      } else if (file.type === Downloadable_Type_Media){
        file.download=file.downloadFrom;
        fe = Media.parse(file);
      }
      //XXX: Temporary assign validate to validateFile
      fe.validateFile = fe.validate;
      return fe;
    };
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="Layout file">
    /**
     * A file of type layout
     */
    function Layout(file){
      var saveAs = file.id+".xlf";
      window.LibraryManager.libFile.call(this,  file.id, saveAs, file.size, Downloadable_Type_Layout);
      this.md5 = file.md5;
      this.updated = undefined;
      this.path = file.path;
      this.isCorrect = undefined;
      this.downloadFrom = file.download;
    }

    Object.assign(Layout.prototype, this.libFile.prototype);

    /**
     * Parse a generic file and return a layout file.
     * @param {Object} file
     * @static
     */
    Layout.parse = function (file) {
      var layout = new Layout(file);
      return layout;
    }

    /**
     *
     * @param successCb
     * @param failureCb
     * @private
     */
    Layout.prototype._downloader = function (successCb, failureCb) {

      if (this.downloadFrom === "http"){
        this.httpDownloader(successCb, failureCb);
      } else {
        var dlInfo = {
          path: this.filename,
          id: this.id,
          size: this.size,
          type: this.type
        };

        var _self = this;

        CommunicationManager.downloadFile(
          function (cbobject) {
            _self.isCorrect = true;
            _self.md5 = _self.md5;
            _self.size = _self.size;
            _self.lastChecked = utils.getTimeStamp();
            _self._validator(
              function () {
                successCb(cbobject);
              },
              function () {
                failureCb(cbObject);
              }
            )

          }, function (cbObject) {
            console.debug("Monitor download layout failed!", cbObject);
            failureCb(cbObject);
          }, dlInfo
        );
      };
    };

    /**
     * Validate the layout
     * @param {Function} cbSuccess Callback if able to validate the file no matter if it is valid or not.
     * @param {Function} bFailure Callback in case an error thrown.
     */
    Layout.prototype._validator = function(cbSuccess, cbFailure) {
      //var options.isCorrect = false;
      var fileOptions = {
        path: this.filename
      };

      var _self = this;

      var checkMD5Cb = function (cbObject) {
        if (cbObject === _self.md5) {
          _self.isCorrect = true;
          cbSuccess({isValid: true});
        } else {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "Wrong MD5!"});
        }
      };

      var checkSizeCb = function (cbObject) {
        if (cbObject.size === parseInt(_self.size)) {
          checkMD5Cb(_self.md5);
        } else {
          /* File exists but has wrong size. Delete it here so it can be downloaded again. */
          var fileOptions = {
            path: _self.filename
          };
          DeviceManager.FileSystem.deleteFile(function () {
            _self.isCorrect = false;
            cbSuccess({
              isValid: false,
              reason: "Wrong file size (expected: " + _self.size + ", got: " + cbObject.size + ")!"
            });
          }, cbFailure, _self.filename);
        }
        // }
      };

      var checkExistCb = function (cbObject) {
        if (cbObject.exists === true) {
          DeviceManager.FileSystem.fileSize.call(_self, checkSizeCb, cbFailure, fileOptions);
        } else {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "File does not exist!"});
        }
      };

      if (_self.manuallyInvalidated) {
        _self.isCorrect = false;
        cbSuccess({isValid: false, reason: "Manually invalidated!"});
      } else {
        DeviceManager.FileSystem.fileExists.call(_self, checkExistCb, cbFailure, fileOptions);
      }
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="Media file">
    /**
     * A file of type media
     */
    function Media(file) {
      //filename to save
      //XXX: Check this against V3 CMS
      var fnsave = (file.download==="xmds") ? file.path : (file.saveAs || file.filename);

      window.LibraryManager.libFile.call(this,  file.id, fnsave, file.size, Downloadable_Type_Media);
      this.md5 = file.md5;
      this.updated = undefined;
      this.path = file.path;
      this.isCorrect = undefined;
      this.downloadFrom = file.download;
    }

    Object.assign(Media.prototype, this.libFile.prototype);

    /**
     * Parse a generic file and return a media file.
     * @param {Object} file
     * @static
     */
    Media.parse = function (file) {
      var media = new Media(file);
      return media;
    }

    Media.prototype._downloader = function (successCb, failureCb) {

      if (this.downloadFrom === "http"){
        this.httpDownloader(successCb, failureCb);
      }else {
        var dlInfo = {
          path: this.filename,
          id: this.id,
          size: this.size,
          type: this.type
        };

        var _self = this;

        CommunicationManager.downloadFile(
          function (cbobject) {
            _self.isCorrect = true;
            // _self.md5 = tf.md5;
            // _self.size = cbobject.size;
            _self.lastChecked = utils.getTimeStamp();
            _self._validator(
              function () {
                if (_self.filename.endsWith(".htz")) {
                  var from = _self.filename;
                  var to = _self.filename.substr(0, _self.filename.length - 4);
                  DeviceManager.FileSystem.unzipFile(
                    function () {
                      successCb(cbobject);
                    },
                    function () {
                    },
                    {from: from, to: to});
                } else {
                  successCb(cbobject);
                }
              },
              function () {
                failureCb(cbObject);
              }
            )

          }, function (cbObject) {
            console.debug("Monitor download media failed!", cbObject);
            failureCb(cbObject);
          }, dlInfo
        );
      }
    };
    Media.prototype._validator = function(cbSuccess, cbFailure){
      //var options.isCorrect = false;
      var fileOptions = {
        path: this.filename
      };

      var _self = this;

      var checkMD5Cb = function (cbObject) {
        if (cbObject === _self.md5) {
          _self.isCorrect = true;
          cbSuccess({isValid: true});
        } else {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "Wrong MD5!"});
        }
      };

      var checkSizeCb = function (cbObject) {
        if (cbObject.size === parseInt(_self.size)) {
          checkMD5Cb(_self.md5);
        } else {
          /* File exists but has wrong size. Delete it here so it can be downloaded again. */
          var fileOptions = {
            path: _self.filename
          };
          DeviceManager.FileSystem.deleteFile(function () {
            _self.isCorrect = false;
            cbSuccess({
              isValid: false,
              reason: "Wrong file size (expected: " + _self.size + ", got: " + cbObject.size + ")!"
            });
          }, cbFailure, _self.filename);
        }
        // }
      };

      var checkExistCb = function (cbObject) {
        if (cbObject.exists === true) {
          DeviceManager.FileSystem.fileSize.call(_self, checkSizeCb, cbFailure, fileOptions);
        } else {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "File does not exist!"});
        }
      };

      if (_self.manuallyInvalidated) {
        _self.isCorrect = false;
        cbSuccess({isValid: false, reason: "Manually invalidated!"});
      } else {
        DeviceManager.FileSystem.fileExists.call(_self, checkExistCb, cbFailure, fileOptions);
      }
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="Dependency file">
    /**
     * A file of type dependency
     *
     * @constructor
     * @param {Object} file
     */
    function Dependency(file) {
      window.LibraryManager.libFile.call(this,  file.id, file.saveAs, file.size, Downloadable_Type_Dependency);
      // this.id = file.id;
      // this.size = file.size;
      // this.saveAs = file.saveAs;
      this.md5 = file.md5;
      this.updated = undefined;
      this.path = file.path;
      this.fileType = file.fileType;
      this.isCorrect = undefined;
      this.downloadFrom = file.download;
    }

    Object.assign(Dependency.prototype, this.libFile.prototype);

    /**
     * Parse a generic file and return a dependency file.
     * @param {Object} file
     * @static
     */
    Dependency.parse = function (file) {
      var dep = new Dependency(file);
      return dep;
    }

    Dependency.prototype._downloader = function (successCb, failureCb) {

      if (this.downloadFrom === "http"){
        this.httpDownloader(successCb, failureCb);
      }else {
        var dlInfo = {
          path: this.filename,
          id: this.id,
          size: this.size,
          type: this.fileType
        };

        var _self = this;
        // CommunicationManager.downloadFile(
        CommunicationManager.getDependency(
          function (cbobject) {
            _self._validator(
              function () {
                successCb(cbobject);
              },
              function () {
                failureCb(cbObject);
              }
            )

          },
          function (cbObject) {
            console.debug("Monitor download file failed!", cbObject);
            failureCb(cbObject);
          },
          dlInfo);
      }
    };

    /**
     * Validate the file
     * @param {Function} cbSuccess Callback if able to validate the file no matter if it is valid or not.
     * @param {Function} bFailure Callback in case an error thrown.
     */
    Dependency.prototype._validator = function(cbSuccess, cbFailure){
      //var options.isCorrect = false;
      var fileOptions = {
        path: this.filename
      };

      var _self = this;

      var checkMD5Cb = function (cbObject) {
        if (cbObject === _self.md5) {
          _self.isCorrect = true;
          cbSuccess({isValid: true});
        } else {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "Wrong MD5!"});
        }
      };

      var checkSizeCb = function (cbObject) {
        if (cbObject.size === parseInt(_self.size)) {
          checkMD5Cb(_self.md5);
        } else {
          /* File exists but has wrong size. Delete it here so it can be downloaded again. */
          var fileOptions = {
            path: _self.filename
          };
          DeviceManager.FileSystem.deleteFile(function () {
            _self.isCorrect = false;
            cbSuccess({
              isValid: false,
              reason: "Wrong file size (expected: " + _self.size + ", got: " + cbObject.size + ")!"
            });
          }, cbFailure, _self.filename);
        }
        // }
      };

      var checkExistCb = function (cbObject) {
        if (cbObject.exists === true) {
          DeviceManager.FileSystem.fileSize.call(_self, checkSizeCb, cbFailure, fileOptions);
        } else {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "File does not exist!"});
        }
      };

      if (_self.manuallyInvalidated) {
        _self.isCorrect = false;
        cbSuccess({isValid: false, reason: "Manually invalidated!"});
      } else {
        DeviceManager.FileSystem.fileExists.call(_self, checkExistCb, cbFailure, fileOptions);
      }
    }
    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="Widget file">
    /**
     * A file of type widget
     *
     * @constructor
     * @param {Object} file
     */
    function Widget(file){
      var saveAs = file.id+".json";
      window.LibraryManager.libFile.call(this,  file.id, saveAs , file.size, Downloadable_Type_Widget);
      this.updateInterval = file.updateInterval;
      this.fileType = file.fileType;
    }

    Object.assign(Widget.prototype, this.libFile.prototype);

    Widget.parse = function (file) {
      var wid = new Widget(file);
      return wid;
    }

    Widget.prototype._downloader = function (successCb, failureCb){
      var dlInfo = {
        id: this.id,
        saveAs: this.filename
      };
      var _self = this;
      CommunicationManager.getData(
        function (cbobject) {

          _self.isCorrect = true;
          _self.manuallyInvalidated = false;
          _self.updated = new Date() + _self.updateInterval;//XXX:Double check that it is ok
          successCb();
        },
        function (cbObject) {
          console.debug("Download widget data file failed!", cbObject);
          failureCb(cbObject);
        },
        dlInfo);
    }

    Widget.prototype._validator = function (cbSuccess, cbFailure, options){
      //var options.isCorrect = false;
      var fileOptions = {
        path: this.filename
      };

      var _self = this;

      var checkUpdatedCb = function (cbObject) {
        if (_self.updated === options.updated) {
          _self.isCorrect = true;
          _self.times = {
            atime: new Date(cbObject.atime),
            mtime: new Date(cbObject.mtime),
            ctime: new Date(cbObject.ctime)
          };
          cbSuccess({isValid: true});
        } else {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "Wrong update time!"});
        }
      };

      var checkExistCb = function (cbObject) {
        if (cbObject.exists === true) {
          DeviceManager.FileSystem.fileSize.call(_self, checkUpdatedCb, cbFailure, fileOptions);
        } else {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "File does not exist!"});
        }
      };

      if (_self.manuallyInvalidated) {
        _self.isCorrect = false;
        cbSuccess({isValid: false, reason: "Manually invalidated!"});
      } else {
        DeviceManager.FileSystem.fileExists.call(_self, checkExistCb, cbFailure, fileOptions);
      }
    }

    // </editor-fold>

    // <editor-fold defaultstate="collapsed" desc="Resource file">

    /**
     * A file of type Resource
     * @param {Object} file
     * @constructor
     */
    function Resource(file) {
      var saveAs = file.id+".html";
      window.LibraryManager.libFile.call(this,  file.id, saveAs , file.size, Downloadable_Type_Resource);
      // this.id = file.id;
      // this.size = file.size;
      // this.saveAs = file.saveAs;
      this.updated = file.updated;
      this.isCorrect = undefined;
      this.layoutid = file.layoutid;
      this.mediaid = file.mediaid;
      this.regionid = file.regionid;
      this.times = {};
      this.fileType = file.fileType;
    }

    Object.assign(Resource.prototype, this.libFile.prototype);

    Resource.parse = function (file) {
      var res = new Resource(file);
      return res;
    }

    Resource.prototype._downloader = function (successCb, failureCb){
      var dlInfo = {
        layoutid: this.layoutid,
        regionid: this.regionid,
        mediaid: this.mediaid,
        saveAs: this.filename

      };

      var _self = this;

      CommunicationManager.downloadResource(function (cbobject) {
        _self.isCorrect = true;
        // this.updated = tf.updated;
        _self.manuallyInvalidated = false;

        _self.validate(
          function (cbobject) {
            successCb();
          },function (cbobject) {
            console.debug("Validate resourse failed!", cbObject, _self);
            failureCb(cbObject);
          },
          {updated: _self.updated}); //XXX: Just pass the same 'updated' value in order to pass the test

      }, function (cbObject) {
        console.debug("Monitor download file failed!", cbObject);
        failureCb(cbObject);
      }, dlInfo);
    }

    Resource.prototype._validator = function (cbSuccess, cbFailure, options){
      //var options.isCorrect = false;
      var fileOptions = {
        path: this.filename
      };

      var _self = this;

      var checkUpdatedCb = function (cbObject) {
        if (_self.updated === options.updated) {
          _self.isCorrect = true;
          _self.times = {
            atime: new Date(cbObject.atime),
            mtime: new Date(cbObject.mtime),
            ctime: new Date(cbObject.ctime)
          };
          cbSuccess({isValid: true});
        } else {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "Wrong update time!"});
        }
      };

      var checkExistCb = function (cbObject) {
        if (cbObject.exists === true) {
          DeviceManager.FileSystem.fileSize.call(_self, checkUpdatedCb, cbFailure, fileOptions);
        } else {
          _self.isCorrect = false;
          cbSuccess({isValid: false, reason: "File does not exist!"});
        }
      };

      if (_self.manuallyInvalidated) {
        _self.isCorrect = false;
        cbSuccess({isValid: false, reason: "Manually invalidated!"});
      } else {
        DeviceManager.FileSystem.fileExists.call(_self, checkExistCb, cbFailure, fileOptions);
      }
    }

    // </editor-fold>
    /* +------------------------+ Private Variables +------------------------+ */
    // <editor-fold defaultstate="collapsed" desc="Private variables">
    //var _self = this;
    var Downloadable_Type_Media = "media";
    var Downloadable_Type_Layout = "layout";
    var Downloadable_Type_Resource = "resource";
    var Downloadable_Type_Dependency = "dependency";
    var Downloadable_Type_Widget = "widget";

    var v_Downloadable_Types = [
      Downloadable_Type_Media,
      Downloadable_Type_Layout,
      Downloadable_Type_Resource,
      Downloadable_Type_Dependency,
      Downloadable_Type_Widget
    ];
    var v_RequiredFiles = [];
    var v_TempRequiredFiles = [];
    var v_blacklisted_widgets = undefined;
    //var v_blacklisted_layouts = undefined;
    // </editor-fold>

    /* +------------------------+ Private functions +------------------------+ */
    // <editor-fold defaultstate="collapsed" desc="Private functions">


    var f_findIndexOfFileByProperty = function (prop, match) {
      var retIndex = null;
      for (var i = 0, max = v_RequiredFiles.length; i < max; i++) {
        if (v_RequiredFiles[i].hasOwnProperty(prop)) {
          if (v_RequiredFiles[i][prop] === match) {
            retIndex = v_RequiredFiles[i];
            break;
          }
        }
      }
      return retIndex;
    };
    /**
     *
     * @param {Object[]} arrProps Array of objects {prop:{any}, match:{any}}
     */
    var f_findIndexOfFileByManyProperties = function (arrProps) {
      var retIndex = null;
      for (var i = 0, max = v_RequiredFiles.length; i < max; i++) {
        var allPropMatch = true;
        var foundOneProp = false;
        for(var j =0, maxp = arrProps.length; j< maxp; j++) {
          if (v_RequiredFiles[i].hasOwnProperty(arrProps[j].prop)) {
            foundOneProp = true;
            allPropMatch &= v_RequiredFiles[i][arrProps[j].prop] === arrProps[j].match;
          }
        }
        if (foundOneProp && allPropMatch){
          retIndex = v_RequiredFiles[i];
          break;
        }
      }
      return retIndex;
    }


    /**
     * @deprecated This will be replaced by checkFileValid.
     * @param {Function} cbSuccess
     * @param {Function} cbFailure
     * @param {object} options
     */
    var checkFile = function (cbSuccess, cbFailure, options) {
      options.isCorrect = false;
      var successSizeCb = function (cbObject) {
        if (cbObject.size === options.size) {
          options.isCorrect = true;
          cbSuccess(true);
        } else {
          /* File exists but has wrong size. Delete it here so it can be downloaded again. */
          var fileOptions = {
            path: options.saveAs
          };
          DeviceManager.FileSystem.deleteFile(function () {
            cbSuccess(false);
          }, cbFailure, options.saveAs);
        }
      };

      var successExistsCb = function (cbObject) {
        if (cbObject.exists === true) {
          if (options.type === "resource") {
            options.isCorrect = true;
            cbSuccess(true);
          } else {
            var fileOptions = {
              path: options.saveAs
            };
            DeviceManager.FileSystem.fileSize(successSizeCb, cbFailure, options.saveAs);
          }
        } else {
          cbSuccess(false);
        }
      };

      var fileOptions = {
        path: options.saveAs
      };

      DeviceManager.FileSystem.fileExists(successExistsCb, cbFailure, options.saveAs);
    };

    var checkFileValid = function (cbSuccess, cbFailure, options) {

      var f = window.LibraryManager.getFileByName(options.filename);
      f.validateFile(cbSuccess, cbFailure, options);

    };

    var checkFileRecursive = function (cbSuccess, cbFailure, options) {
      options.v_fileIndex++;

      var checkFileSuccessCb = function (cbObject) {
        if (cbObject === false) {
          options.v_needsUpdate = true;
        }
        checkFileRecursive(cbSuccess, cbFailure, options);
      };

      if (options.v_fileIndex >= v_RequiredFiles.length) {
        /* Finish going through files */
        cbSuccess({needsUpdate: options.v_needsUpdate});

      } else {
        //if (v_RequiredFiles[options.v_fileIndex].hasOwnProperty("path")){
        if (v_RequiredFiles[options.v_fileIndex].hasOwnProperty("type")) {
          if (LibraryManager.isDownloadable(v_RequiredFiles[options.v_fileIndex])) {
            /* Add time stamp of check */
            v_RequiredFiles[options.v_fileIndex].lastChecked = utils.getTimeStamp();

            if (v_RequiredFiles[options.v_fileIndex].type === "resource") {
              v_RequiredFiles[options.v_fileIndex].path = /*"resource_" + */ v_RequiredFiles[options.v_fileIndex].id;
              v_RequiredFiles[options.v_fileIndex].saveAs = v_RequiredFiles[options.v_fileIndex].path;
            }

            checkFile(checkFileSuccessCb, cbFailure, v_RequiredFiles[options.v_fileIndex]);
          } else {
            checkFileRecursive(cbSuccess, cbFailure, options);
          }
        } else {
          checkFileRecursive(cbSuccess, cbFailure, options);
        }
      }
    };

    // </editor-fold>

    /* +------------------------+ Public functions +------------------------+ */
    // <editor-fold defaultstate="collapsed" desc="Public functions">

    this.setRequiredFiles = function (files) {
      v_TempRequiredFiles = files;
    };

    this.getRequiredFiles = function () {
      //return (JSON.parse(JSON.stringify(v_RequiredFiles))); /* Return a copy */
      //return v_RequiredFiles;
      return v_TempRequiredFiles;
    };


    this.checkFilesOK = function () {
      for (var i = 0, max = v_RequiredFiles.length; i < max; i++) {
        if (v_RequiredFiles[i].isCorrect === false) {
          return false;
        }
      }

      return true;

    };

    /**
     * Check a file if exists and has the correct size in filesystem.
     * @param {type} cbSuccess Callback function when check completes.
     * @param {type} cbFailure Callback function in case of an error.
     * @param {type} file The file to examine.
     * @returns None
     * @see It will update its isCorrect field appropriately.
     */
    this.checkFileOK = function (cbSuccess, cbFailure, file) {
      checkFileValid(cbSuccess, cbFailure, file);
    };

    /**
     * Add a new file in library.
     * @param {type} cbSuccess Callback function with a file Object checked for validity.
     * @param {type} cbFailure Callback function in case of a failure.
     * @param {Object} file File should contain filename, size and type fields.
     * @returns None
     * @see Adding a file it will automatically detect if file is in file system and if it is correct!
     */
    this.addFile = function (cbSuccess, cbFailure, file) {
      var tf = {};
      var fe = f_findIndexOfFileByManyProperties([
        {"prop":"id", "match":file.id},
        {"prop":"type", "match":file.type},
        {"prop":"fileType", "match":file.fileType}
      ]);

      // If file exists and has the path property, update it's path.
      if (file.path && fe !== null) {
        fe.path = file.path;
      }

      if (file.type === Downloadable_Type_Dependency) {
        //var fe = LibraryManager.getFileByName(file.path);
        if (fe === null) {
          fe = Dependency.parse(file);
          v_RequiredFiles.push(fe);

          //XXX: Temporary assign validate to validateFile
          fe.validateFile = fe.validate;
        }
        // else {
        //   fe.id = file.id;
        // }
      } else if (file.type === Downloadable_Type_Resource ){
        // var fe = f_findIndexOfFileByManyProperties([
        //   {"prop":"id", "match":file.id},
        //   {"prop":"type", "match":file.type}
        // ]);
        //var fe = LibraryManager.getFileById(file.id);
        if (fe === null) {
          fe = Resource.parse(file);
          v_RequiredFiles.push(fe);

          //XXX: Temporary assign validate to validateFile
          fe.validateFile = fe.validate;
        }
        // else {
        //   fe.id = file.id;
        // }
        tf.updated = fe.updated;
      } else if (file.type === Downloadable_Type_Widget ){
        var fileName = file.id+".json";
        var fe = window.LibraryManager.getFileByName(fileName);
        if (fe === null) {
          fe = Widget.parse(file);
          v_RequiredFiles.push(fe);

          //XXX: Temporary assign validate to validateFile
          fe.validateFile = fe.validate;
        }
        // else {
        //   fe.id = file.id;
        // }
        tf.updated = fe.updated;
      } else if (file.type === Downloadable_Type_Layout ){
        // var fe = LibraryManager.getFileById(file.id);
        if (fe === null) {
          fe = Layout.parse(file);
          v_RequiredFiles.push(fe);

          //XXX: Temporary assign validate to validateFile
          fe.validateFile = fe.validate;
        }
        // else {
        //   fe.id = file.id;
        // }
        tf.updated = fe.updated;
      } else if (file.type === Downloadable_Type_Media){
        // var fe = f_findIndexOfFileByManyProperties([
        //   {"prop":"id", "match":file.id},
        //   {"prop":"type", "match":file.type}
        // ]);
        // var fe = LibraryManager.getFileById(file.id);
        if (fe === null) {
          fe = Media.parse(file);
          v_RequiredFiles.push(fe);

          //XXX: Temporary assign validate to validateFile
          fe.validateFile = fe.validate;
        }
        // else {
        //   fe.id = file.id;
        // }
        tf.updated = fe.updated;
      } else {
        console.warn("Add file is covered from the general case!");
        switch (file.download) {
          case "http":
            tf = {
              id: file.id,
              filename: file.saveAs,
              size: file.size,
              type: file.type,
              md5: file.md5
            };

            if (!defined(tf.filename)) {
              switch (tf.type) {
                case "layout":
                  tf.filename = file.id + ".xlf";
                  break;
                default:
                  throw ERRORS.ParameterNotDefined();
                //break;
              }
            }

            fe = this.getFileByName(tf.filename);
            if (fe === null) {
              fe = new this.libFile(tf.id, tf.filename, tf.size, tf.type);
              fe.md5 = tf.md5;
              file.code && (fe.code = file.code);
              v_RequiredFiles.push(fe);

            } else {
              fe.id = tf.id;
            }
            fe.downloader = function (successCb, failureCb) {

              var dlInfo = {
                source: file.path,
                destination: tf.filename
              };
              DeviceManager.FileSystem.downloadFile(function () {
                fe.isCorrect = true;
                fe.md5 = tf.md5;
                fe.size = tf.size;
                fe.validateFile(function (cbobject) {
                  fe.lastChecked = utils.getTimeStamp();
                  if (cbobject.hasOwnProperty('isValid')) {
                    if (!cbobject.isValid) {
                      Logger.cmsLog({
                        category: "audit",
                        message: "Validating media " + fe.filename + " failed. Reason: " + cbobject.reason,
                        mediaID: fe.id,
                        method: "validateFile",
                        thread: "LibraryManager.addFile.downloader"
                      });
                    }
                  }
                  if (fe.filename.endsWith(".htz")) {
                    var from = fe.filename;
                    var to = fe.filename.substr(0, fe.filename.length - 4);
                    DeviceManager.FileSystem.unzipFile(
                      function () {
                        successCb(cbobject);
                      },
                      function () {
                      },
                      {from: from, to: to});
                  } else {
                    successCb(cbobject);
                  }
                }, function (cbObj) {
                  Logger.cmsLog({
                    category: "audit",
                    message: "Download file failed (not valid). Reason: " + JSON.stringify(cbObj),
                    method: "validateFile",
                    thread: "LibraryManager.addFile.downloader"
                  });
                  failureCb(cbObj);
                }, fe);
              }, function (cbObject) {
                var message = JSON.stringify(cbObject);
                if (cbObject.httpStatus && cbObject.httpStatus === 404) {
                  message = "File not found. Widget id: " + fe.id;
                }
                ;
                Logger.cmsLog({
                  category: "audit",
                  message: "Monitor download file failed. Reason: " + message,
                  method: "validateFile",
                  thread: "LibraryManager.addFile.downloader"
                });
                failureCb(cbObject);
              }, dlInfo);
            };
            break;
          case "xmds":
            tf = {
              id: file.id,
              size: file.size,
              type: file.type,
              updated: file.updated,
              md5: file.md5
            };

            switch (file.type) {
              case "layout":
                tf.filename = file.id + ".xlf";
                fe = this.getFileByName(tf.filename);
                if (fe === null) {
                  fe = new this.libFile(tf.id, tf.filename, tf.size, tf.type);
                  fe.md5 = tf.md5;
                  v_RequiredFiles.push(fe);
                } else {
                  fe.id = tf.id;
                }

                fe.downloader = function (successCb, failureCb) {

                  var dlInfo = {
                    path: fe.filename,
                    id: fe.id,
                    size: fe.size,
                    type: fe.type
                  };

                  CommunicationManager.downloadFile(
                    function (cbobject) {
                      fe.isCorrect = true;
                      fe.md5 = tf.md5;
                      fe.size = tf.size;
                      fe.lastChecked = utils.getTimeStamp();
                      successCb(cbobject);

                    }, function (cbObject) {
                      console.debug("Monitor download file failed!", cbObject);
                      failureCb(cbObject);
                    }, dlInfo);
                };
                break;
              case "media":
                tf.filename = file.path;
                fe = this.getFileByName(tf.filename);
                if (fe === null) {
                  fe = new this.libFile(tf.id, tf.filename, tf.size, tf.type);
                  fe.md5 = tf.md5;
                  v_RequiredFiles.push(fe);
                } else {
                  fe.id = tf.id;
                }

                fe.downloader = function (successCb, failureCb) {

                  var dlInfo = {
                    path: fe.filename,
                    id: fe.id,
                    size: fe.size,
                    type: fe.type
                  };

                  CommunicationManager.downloadFile(
                    function (cbobject) {
                      fe.isCorrect = true;
                      fe.md5 = tf.md5;
                      fe.size = tf.size;
                      successCb(cbobject);
                    }, function (cbObject) {
                      console.debug("Monitor download file failed!", cbObject);
                      failureCb(cbObject);
                    }, dlInfo);
                };
                break;
              case Downloadable_Type_Dependency:
                var fe = window.LibraryManager.getFileByName(file.path);
                if (fe === null) {
                  fe = Dependency.parse(file);
                  v_RequiredFiles.push(fe);

                  //XXX: Temporary assign validate to validateFile
                  fe.validateFile = fe.validate;
                } else {
                  fe.id = file.id;
                }
                break;
              default:
                console.warn("Could not download " + file.type + " from XMDS.");
                break;
            }
            break;
          default: //All other cases (Resources)
            tf = {
              filename: file.id + ".html",
              id: file.id,
              layoutid: file.layoutid,
              mediaid: file.mediaid,
              regionid: file.regionid,
              type: file.type,
              updated: file.updated
            };

            fe = this.getFileByName(tf.filename);
            if (fe === null) {
              fe = new this.libFile(tf.id, tf.filename, tf.size, tf.type);
              fe.updated = tf.updated;
              v_RequiredFiles.push(fe);
            } else {
              fe.id = tf.id;
            }
            fe.downloader = function (successCb, failureCb) {

              var dlInfo = {
                layoutid: tf.layoutid,
                regionid: tf.regionid,
                mediaid: tf.mediaid,
                saveAs: tf.filename

              };
              CommunicationManager.downloadResource(function (cbobject) {
                fe.isCorrect = true;
                fe.updated = tf.updated;
                fe.manuallyInvalidated = false;

                /*****************************************************************************************************/
                fe.validateFile(function (cbobject) {
                  if (cbobject.hasOwnProperty('isValid')) {
                    if (!cbobject.isValid) {
                      Logger.cmsLog({
                        category: "audit",
                        message: "Validating resource " + fe.filename + " failed. Reason: " + cbobject.reason,
                        mediaID: fe.id,
                        method: "validateFile",
                        thread: "LibraryManager.addFile.downloader"
                      });
                    }
                  }
                  successCb(cbobject);
                }, function (cbObj) {
                  Logger.cmsLog({
                    category: "audit",
                    message: "Download resource failed (not valid). Reason: " + JSON.stringify(cbObj),
                    method: "validateFile",
                    thread: "LibraryManager.addFile.downloader"
                  });
                  failureCb(cbObj);
                }, fe);
                /*****************************************************************************************************/

              }, function (cbObject) {
                console.debug("Monitor download file failed!", cbObject);
                failureCb(cbObject);
              }, dlInfo);
            };


            break;
        }
      }
      if (defined(fe)) {
        fe.validateFile(function (cbObj) {
          //console.debug("File validated : " + f.filename, cbObj);
          if (!fe.isCorrect) {
            console.debug("File " + fe.filename + " not valid. Reason: " + cbObj.reason);
          }
          cbSuccess(fe);
        }, function (e) {
          Logger.cmsFault(undefined, undefined,
            {
              code: ERRORS.FaultCodes.RemoteResourceFailed,
              reason: "Invalid file : " + fe.filename,
              type: "ERROR",
              method: "addFile",
              thread: "LibraryManager",
              date: new Date(),
              expires: new Date((new Date()).valueOf() + DeviceManager.Config.getOptions().collectInterval * 1000)
            }
          );
          //console.debug("File validation failed :" + fe.filename,cbObj);
        }, fe);
      } else {
        cbSuccess(null);
      }
    };

    /**
     * Get file by name (path)
     * @param {String} name The name of the file (actually the path).
     * @returns {libFile} The file. Null if file not exists.
     */
    this.getFileByName = function (name) {
      return f_findIndexOfFileByProperty("filename", name);
    };

    /**
     * Get file by index.
     * @param {type} index The index of the file.
     * @returns {Object} The file.
     */
    this.getFileByIndex = function (index) {
      if (index >= v_RequiredFiles.length) {
        return null;
      }

      return v_RequiredFiles[index];
    };

    /**
     * Get file by id.
     * @param {String} id The id number.
     * @returns {Object} The file.
     */
    this.getFileById = function (id) {
      var idx = f_findIndexOfFileByProperty("id", id);

      return idx;
    };

    /**
     * Get file by action code.
     * @param {String} code The code.
     * @returns {Object} The file.
     */
    this.getFileByActionCode = function (code) {
      var idx = f_findIndexOfFileByProperty("code", code);
      return idx;
    };
    this.getFileByMultipleProperties = function(props){
      var idx = f_findIndexOfFileByManyProperties(props);
      return idx;
    }

    /*
     * Remove a file from library.
     * @param {Function} cbSuccess Callback function called after file deleted succesfull.
     * @param {Function} cbFailure Callback function called in case of a failure.
     * @param {String} filename The filename of the file.
     * @returns None
     * @see It will also remove the file from the filesystem.
     */
    this.removeFile = function (cbSuccess, cbFailure, filename) {
      //var f = LibraryManager.getFileByName(filename);

      var idx = null;
      for (var i = 0, max = v_RequiredFiles.length; i < max; i++) {
        if (v_RequiredFiles[i].filename === filename) {
          idx = i;
          break;
        }
      }
      if (idx !== null) {
        v_RequiredFiles.splice(idx, 1);
      }
      DeviceManager.FileSystem.deleteFile(
        cbSuccess,
        function (cbObject) {
          console.debug("Error", cbObject, filename);
          //if (cbObject.errorText.substring(0,20) === "File does not exists")
          if (cbObject.code === 8) //File is not avalaible
          {
            cbSuccess();
          } else {
            cbFailure(cbObject);
          }
        },
        filename
      );

    };

    /**
     * Empty the files list.
     * @version 1.8.1
     * @since 1.8.1
     */
    this.clearAllFiles = function () {
      v_RequiredFiles = [];
    };

    /**
     * Returns the count of files.
     * @returns {Number} Count of files.
     */
    this.length = function () {
      return v_RequiredFiles.length;
    };

    /**
     * Check if a file can be downloaded from CMS.
     * @param {type} filename The name of a file.
     * @returns {Boolean} True if it can else false.
     */
    this.isDownloadable = function (filename) {
      if (!defined(filename)) {
        return false;
      }
      var f = this.getFileByName(filename);
      return v_Downloadable_Types.indexOf(f.type) !== -1;
    };

    /**
     * Check if a file is valid.
     * @param {String} filename The name of a file.
     * @returns {Boolean} True if it is valid else false.
     */
    this.isFileOk = function (filename) {
      if (!defined(filename)) {
        return false;
      }
      var f = this.getFileByName(filename);

      if (f === null) {
        console.debug('LibraryManager.libFile.isFileOk: ' + filename + ' not in v_RequiredFiles!');
        return false;
      }

      var r = f.isCorrect;

      r = r && (typeof f.isBlacklisted !== 'undefined' ? !f.isBlacklisted : true);

      return r;
    };

    /**
     * Saves the library.
     * @returns None
     */
    this.saveLibrary = function () {
      localStorage.library = JSON.stringify(v_RequiredFiles);
    };

    /**
     * Loads the library.
     * @returns {Boolean} True if library loaded successfully else false.
     */
    this.loadLibrary = function () {
      if (localStorage.hasOwnProperty("library")) {
        var tmplibfiles = [];
        try {
          tmplibfiles = JSON.parse(localStorage.library);
        } catch (e) {
          console.warn('LibraryManager.loadLibrary: Parse error', e.message, e.stack);
          return false;
        }
        try {
          for (var i = 0, max = tmplibfiles.length; i < max; i++) {
            var f;
              // f = new this.libFile(tmplibfiles[i].id, tmplibfiles[i].filename, tmplibfiles[i].size, tmplibfiles[i].type, tmplibfiles[i].download);
            f = this.libFile.factoryFromLocalStorage(tmplibfiles[i]);
            f.filename = tmplibfiles[i].filename;
            f.md5 = tmplibfiles[i].md5;
            f.isCorrect = tmplibfiles[i].isCorrect;
            tmplibfiles[i].code && (f.code = tmplibfiles[i].code);
            if (tmplibfiles[i].type === "resource") {
              f.updated = tmplibfiles[i].updated;
            }
            v_RequiredFiles.push(f);
          }
        } catch (e) {
          console.warn('LibraryManager.loadLibrary: Error', e.message, e.stack);
          return false;
        }
      } else {
        console.debug('LibraryManager.loadLibrary: No library in localStorage');
      }
      return true;
    };

    /**
     * Clean up the library from old files
     * @param {Function} cbSuccess
     * @param {Function} cbFailure
     * @returns None
     */
    this.cleanUpLibrary = function (cbSuccess, cbFailure) {
      var i, j, maxi, maxj;
      console.debug("Clean up library");
      var markRemove = [];
      for (i = 0; i < v_RequiredFiles.length; i++) {
        var found = false;
        for (var j = 0; j < v_TempRequiredFiles.length; j++) {
          if (v_TempRequiredFiles[j].hasOwnProperty("type")) {
            if (LibraryManager.isCompatibleType(v_TempRequiredFiles[j].type)) {
              if (v_TempRequiredFiles[j].download === "xmds") {
                if (v_TempRequiredFiles[j].type === Downloadable_Type_Layout) {
                  v_TempRequiredFiles[j].saveAs = v_TempRequiredFiles[j].id + ".xlf";
                } else if (v_TempRequiredFiles[j].type === Downloadable_Type_Media) {
                  v_TempRequiredFiles[j].saveAs = v_TempRequiredFiles[j].path;
                }
              }else{
                if (v_TempRequiredFiles[j].type === Downloadable_Type_Resource) {
                  v_TempRequiredFiles[j].saveAs = v_TempRequiredFiles[j].id +".html";
                } else if (v_TempRequiredFiles[j].type === Downloadable_Type_Widget) {
                  v_TempRequiredFiles[j].saveAs = v_TempRequiredFiles[j].id +".json";
                }
              }
              if (v_RequiredFiles[i].filename === v_TempRequiredFiles[j].saveAs) {
                found = true;
                break;
              }
            } else if (v_TempRequiredFiles[j].type === Downloadable_Type_Resource) {
              console.info("That should be resolved from the previous cases");
              if (v_RequiredFiles[i].filename === v_TempRequiredFiles[j].id + ".html") {
                found = true;
                break;
              }

            }
          }
        }
        if (!found) {
          markRemove.push(v_RequiredFiles[i].filename);
        }
      }


      var itemsRemoved = false;

      var removeEach = function () {
        if (markRemove.length > 0) {
          itemsRemoved = true;
          window.LibraryManager.removeFile(removeEach, cbFailure, markRemove.pop());
        } else {

          cbSuccess(itemsRemoved);
        }
      };

      /* Issue #21 */
      DeviceManager.FileSystem.listFiles(
        function (files) {
          for (i = 0, maxi = files.length; i < maxi; i++) {
            if (files[i].isFile) {
              var found = false;
              for (j = 0, maxj = v_RequiredFiles.length; j < maxj; j++) {
                //if (encodeURIComponent(v_RequiredFiles[j].filename) === files[i].name) {
                if (v_RequiredFiles[j].filename === files[i].name) {
                  found = true;
                  break;
                }
              }
              if (!found) {
                console.debug("Delete:" + files[i].name);
                markRemove.push(files[i].name);
              }
            }
          }
          removeEach();
        },
        function (cbObject) {
          console.debug("Could not list files!", cbObject);
          removeEach();
        }, null
      );

    };


    /**
     * Update file metadata (after download)
     * @param {Object} file The file downloaded
     * @returns {undefined}
     * @throws {ERRORS.ParameterNotDefined} If file has no filename.
     * @throws {ERRORS.FileNotFound} If file is not found in library.
     */
    this.updateFile = function (file) {
      if (!defined(file.saveAs)) {
        throw ERRORS.ParameterNotDefined("Filename is not defined in file.");
      }

      var fe = this.getFileByName(file.saveAs);

      if (fe === null) {
        throw ERRORS.FileNotFound();
      }

      if (fe.type !== "resource") {
        fe.size = file.size;
        fe.type = file.type;
        fe.md5 = file.md5;

      } else {
        fe.type = file.type;
        fe.updated = file.updated;
      }

      window.LibraryManager.saveLibrary();
    };


    this.getLibraryFiles = function () {
      return v_RequiredFiles;
    };

    /**
     * Array of blacklisted widgets
     * @property {Boolean} LibraryManager.blacklisted_widgets .
     * @version 1.8.18
     * @since 1.8.1
     */
    Object.defineProperty(this, 'blacklisted_widgets', {
      get: function () {
        if (!v_blacklisted_widgets) {
//	                if (localStorage.hasOwnProperty('blacklist_widgets')){
//	                    v_blacklisted_widgets = JSON.parse(localStorage.blacklist_widgets);
//	                    application.statusPage.blacklist(v_blacklisted_widgets);
//	                } else {
          v_blacklisted_widgets = [];
//	                }
          var localPush = v_blacklisted_widgets.push;
          v_blacklisted_widgets.push = function (value) {
            if (v_blacklisted_widgets.indexOf(value) === -1) {
              localPush.call(v_blacklisted_widgets, value);
//	                        localStorage.blacklist_widgets = JSON.stringify(v_blacklisted_widgets);
              application.statusPage.blacklist(v_blacklisted_widgets);
              var lfile = window.LibraryManager.getFileById(value);
              if (lfile !== null) {
                lfile.isBlacklisted = true;
                ScheduleManager.revalidateContent();
              }
            }
          };
        }

        return v_blacklisted_widgets;
      }
    });

  /**
   * Check if type is downloadable by the library handler.
   * @param {String} type The type top check.
   * @returns {Boolean} True if it is otherwise false.
   */
  this.isCompatibleType = function(type){
    return v_Downloadable_Types.indexOf(type) >= 0;
  };
    var fnLoadLib = this.loadLibrary.bind(this);
    setTimeout(
      fnLoadLib,
      0
    );

    // </editor-fold>
  }();
}());
